@@ -87,7 +87,7 @@ pub(crate) struct OperatorValidator {
 
     // This is a list of flags for wasm features which are used to gate various
     // instructions.
-    features: WasmFeatures,
+    pub(crate) features: WasmFeatures,
 }
 
 // This structure corresponds to `ctrl_frame` as specified at in the validation
@@ -353,6 +353,11 @@ impl OperatorValidator {
                 "alignment must not be larger than natural",
             ));
         }
+        if index_ty == Type::I32 && memarg.offset > u64::from(u32::MAX) {
+            return Err(OperatorValidatorError::new(
+                "offset out of range: must be <= 2**32",
+            ));
+        }
         Ok(index_ty)
     }
 
@@ -604,7 +609,7 @@ impl OperatorValidator {
                     unreachable: false,
                 });
                 // Push exception argument types.
-                let ty = event_at(&resources, index)?;
+                let ty = tag_at(&resources, index)?;
                 for ty in ty.inputs() {
                     self.push_operand(ty)?;
                 }
@@ -612,7 +617,7 @@ impl OperatorValidator {
             Operator::Throw { index } => {
                 self.check_exceptions_enabled()?;
                 // Check values associated with the exception.
-                let ty = event_at(&resources, index)?;
+                let ty = tag_at(&resources, index)?;
                 for ty in ty.inputs().rev() {
                     self.pop_operand(Some(ty))?;
                 }
@@ -1840,10 +1845,16 @@ impl OperatorValidator {
                 self.check_bulk_memory_enabled()?;
                 let dst_ty = self.check_memory_index(dst, resources)?;
                 let src_ty = self.check_memory_index(src, resources)?;
+
+                // The length operand here is the smaller of src/dst, which is
+                // i32 if one is i32
                 self.pop_operand(Some(match src_ty {
                     Type::I32 => Type::I32,
                     _ => dst_ty,
                 }))?;
+
+                // ... and the offset into each memory is required to be
+                // whatever the indexing type is for that memory
                 self.pop_operand(Some(src_ty))?;
                 self.pop_operand(Some(dst_ty))?;
             }
@@ -1972,13 +1983,10 @@ fn func_type_at<T: WasmModuleResources>(
         .ok_or_else(|| OperatorValidatorError::new("unknown type: type index out of bounds"))
 }
 
-fn event_at<T: WasmModuleResources>(
-    resources: &T,
-    at: u32,
-) -> OperatorValidatorResult<&T::FuncType> {
+fn tag_at<T: WasmModuleResources>(resources: &T, at: u32) -> OperatorValidatorResult<&T::FuncType> {
     resources
-        .event_at(at)
-        .ok_or_else(|| OperatorValidatorError::new("unknown event: event index out of bounds"))
+        .tag_at(at)
+        .ok_or_else(|| OperatorValidatorError::new("unknown tag: tag index out of bounds"))
 }
 
 enum Either<A, B> {
