@@ -114,7 +114,7 @@ pub enum SectionCode<'a> {
     Code,       // Function bodies (code)
     Data,       // Data segments
     DataCount,  // Count of passive data segments
-    Event,      // Event declarations
+    Tag,        // Tag declarations
 }
 
 /// Types as defined [here].
@@ -155,7 +155,7 @@ pub enum ExternalKind {
     Function,
     Table,
     Memory,
-    Event,
+    Tag,
     Global,
     Type,
     Module,
@@ -193,45 +193,54 @@ pub struct ExportType<'a> {
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub struct ResizableLimits {
+pub struct TableType {
+    pub element_type: Type,
+    /// Initial size of this table, in elements.
     pub initial: u32,
+    /// Optional maximum size of the table, in elements.
     pub maximum: Option<u32>,
 }
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub struct ResizableLimits64 {
+pub struct MemoryType {
+    /// Whether or not this is a 64-bit memory, using i64 as an index. If this
+    /// is false it's a 32-bit memory using i32 as an index.
+    ///
+    /// This is part of the memory64 proposal in WebAssembly.
+    pub memory64: bool,
+
+    /// Whether or not this is a "shared" memory, indicating that it should be
+    /// send-able across threads and the `maximum` field is always present for
+    /// valid types.
+    ///
+    /// This is part of the threads proposal in WebAssembly.
+    pub shared: bool,
+
+    /// Initial size of this memory, in wasm pages.
+    ///
+    /// For 32-bit memories (when `memory64` is `false`) this is guaranteed to
+    /// be at most `u32::MAX` for valid types.
     pub initial: u64,
+
+    /// Optional maximum size of this memory, in wasm pages.
+    ///
+    /// For 32-bit memories (when `memory64` is `false`) this is guaranteed to
+    /// be at most `u32::MAX` for valid types. This field is always present for
+    /// valid wasm memories when `shared` is `true`.
     pub maximum: Option<u64>,
 }
 
-#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub struct TableType {
-    pub element_type: Type,
-    pub limits: ResizableLimits,
-}
-
-#[derive(Debug, Copy, Clone, PartialEq, Eq, Hash)]
-pub enum MemoryType {
-    M32 {
-        limits: ResizableLimits,
-        shared: bool,
-    },
-    M64 {
-        limits: ResizableLimits64,
-        shared: bool,
-    },
-}
-
 #[derive(Debug, Copy, Clone, PartialEq, Eq)]
-pub struct EventType {
+pub struct TagType {
     pub type_index: u32,
 }
 
 impl MemoryType {
     pub fn index_type(&self) -> Type {
-        match self {
-            MemoryType::M32 { .. } => Type::I32,
-            MemoryType::M64 { .. } => Type::I64,
+        if self.memory64 {
+            Type::I64
+        } else {
+            Type::I32
         }
     }
 }
@@ -247,7 +256,7 @@ pub enum ImportSectionEntryType {
     Function(u32),
     Table(TableType),
     Memory(MemoryType),
-    Event(EventType),
+    Tag(TagType),
     Global(GlobalType),
     Module(u32),
     Instance(u32),
@@ -257,7 +266,20 @@ pub enum ImportSectionEntryType {
 pub struct MemoryImmediate {
     /// Alignment, stored as `n` where the actual alignment is `2^n`
     pub align: u8,
-    pub offset: u32,
+
+    /// A fixed byte-offset that this memory immediate specifies.
+    ///
+    /// Note that the memory64 proposal can specify a full 64-bit byte offset
+    /// while otherwise only 32-bit offsets are allowed. Once validated
+    /// memory immediates for 32-bit memories are guaranteed to be at most
+    /// `u32::MAX` whereas 64-bit memories can use the full 64-bits.
+    pub offset: u64,
+
+    /// The index of the memory this immediate points to.
+    ///
+    /// Note that this points within the module's own memory index space, and
+    /// is always zero unless the multi-memory proposal of WebAssembly is
+    /// enabled.
     pub memory: u32,
 }
 
@@ -272,6 +294,13 @@ pub enum NameType {
     Module,
     Function,
     Local,
+    Label,
+    Type,
+    Table,
+    Memory,
+    Global,
+    Element,
+    Data,
     Unknown(u32),
 }
 
